@@ -4,7 +4,7 @@
 #
 # Copyright (C) 2004 Robert Kaye
 # Copyright (C) 2006-2009, 2011-2013, 2017 Lukáš Lalinský
-# Copyright (C) 2007-2011, 2015, 2018-2020 Philipp Wolfer
+# Copyright (C) 2007-2011, 2015, 2018-2022 Philipp Wolfer
 # Copyright (C) 2008 Gary van der Merwe
 # Copyright (C) 2008-2009 Nikolai Prokoschenko
 # Copyright (C) 2009 Carlin Mangar
@@ -16,15 +16,17 @@
 # Copyright (C) 2012-2014 Wieland Hoffmann
 # Copyright (C) 2013 Calvin Walton
 # Copyright (C) 2013-2014 Ionuț Ciocîrlan
-# Copyright (C) 2013-2014, 2017 Sophist-UK
-# Copyright (C) 2013-2014, 2017-2019 Laurent Monin
+# Copyright (C) 2013-2014, 2017, 2021 Sophist-UK
+# Copyright (C) 2013-2014, 2017-2022 Laurent Monin
 # Copyright (C) 2016 Rahul Raturi
 # Copyright (C) 2016 Ville Skyttä
 # Copyright (C) 2016-2018 Sambhav Kothari
 # Copyright (C) 2017-2018 Antonio Larrosa
 # Copyright (C) 2019 Joel Lintunen
 # Copyright (C) 2020 Ray Bouchard
-# Copyright (C) 2020 Gabriel Ferreira
+# Copyright (C) 2020-2021 Gabriel Ferreira
+# Copyright (C) 2021 Bob Swift
+# Copyright (C) 2021 Petit Minion
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -41,14 +43,17 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 
-from collections import defaultdict
+from collections import Counter
+from enum import (
+    Enum,
+    auto,
+)
 import fnmatch
 from functools import partial
 import os
 import os.path
 import re
 import shutil
-import unicodedata
 
 from PyQt5 import QtCore
 
@@ -57,7 +62,6 @@ from picard import (
     log,
 )
 from picard.config import get_config
-from picard.const import QUERY_LIMIT
 from picard.const.sys import (
     IS_MACOS,
     IS_WIN,
@@ -70,17 +74,21 @@ from picard.plugin import (
     PluginFunctions,
     PluginPriority,
 )
+from picard.script import get_file_naming_script
 from picard.util import (
+    any_exception_isinstance,
     decode_filename,
     emptydir,
     find_best_match,
     format_time,
     is_absolute_path,
-    samefile,
+    normpath,
     thread,
-    tracknum_from_filename,
+    tracknum_and_title_from_filename,
 )
 from picard.util.filenaming import (
+    get_available_filename,
+    make_save_path,
     make_short_filename,
     move_ensure_casing,
 )
@@ -91,7 +99,11 @@ from picard.util.tags import PRESERVED_TAGS
 from picard.ui.item import Item
 
 
-FILE_INFO_TAGS = ('~bitrate', '~sample_rate', '~channels', '~bits_per_sample', '~format')
+class FileErrorType(Enum):
+
+    UNKNOWN = auto()
+    NOTFOUND = auto()
+    NOACCESS = auto()
 
 
 class File(QtCore.QObject, Item):
@@ -110,6 +122,10 @@ class File(QtCore.QObject, Item):
     LOOKUP_METADATA = 1
     LOOKUP_ACOUSTID = 2
 
+    EXTENSIONS = []
+
+    FILE_INFO_TAGS = ('~bitrate', '~sample_rate', '~channels', '~bits_per_sample', '~format')
+
     comparison_weights = {
         "title": 13,
         "artist": 4,
@@ -139,6 +155,7 @@ class File(QtCore.QObject, Item):
         self.base_filename = os.path.basename(filename)
         self._state = File.UNDEFINED
         self.state = File.PENDING
+        self.error_type = FileErrorType.UNKNOWN
 
         self.orig_metadata = Metadata()
         self.metadata = Metadata()
@@ -156,9 +173,43 @@ class File(QtCore.QObject, Item):
     def __repr__(self):
         return '<%s %r>' % (type(self).__name__, self.base_filename)
 
-    @property
-    def new_metadata(self):
-        return self.metadata
+    # pylint: disable=no-self-use
+    def format_specific_metadata(self, metadata, tag, settings=None):
+        """Can be overridden to customize how a tag is displayed in the UI.
+        This is useful if a tag saved to the underlying format will differ from
+        the internal representation in a way that would cause data loss. This is e.g.
+        the case for some ID3v2.3 tags.
+
+        Args:
+            metadata: The metadata object to read the tag from
+            tag: Name of the tag
+            settings: Dictionary of settings. If not set, config.setting should be used
+
+        Returns:
+            An array of values for the tag
+        """
+        return metadata.getall(tag)
+
+    def _format_specific_copy(self, metadata, settings=None):
+        """Creates a copy of metadata, but applies format_specific_metadata() to the values.
+        """
+        copy = Metadata(
+            deleted_tags=metadata.deleted_tags,
+            images=metadata.images,
+            length=metadata.length)
+        for name in metadata:
+            copy[name] = self.format_specific_metadata(metadata, name, settings)
+        return copy
+
+    def _set_error(self, error):
+        self.state = File.ERROR
+        self.error_append(str(error))
+        if any_exception_isinstance(error, FileNotFoundError):
+            self.error_type = FileErrorType.NOTFOUND
+        elif any_exception_isinstance(error, PermissionError):
+            self.error_type = FileErrorType.NOACCESS
+        else:
+            self.error_type = FileErrorType.UNKNOWN
 
     def load(self, callback):
         thread.run_task(
@@ -184,9 +235,9 @@ class File(QtCore.QObject, Item):
     def _loading_finished(self, callback, result=None, error=None):
         if self.state != File.PENDING or self.tagger.stopping:
             return
+        config = get_config()
         if error is not None:
-            self.state = self.ERROR
-            self.error_append(str(error))
+            self._set_error(error)
 
             # If loading failed, force format guessing and try loading again
             from picard.formats.util import guess_format
@@ -214,9 +265,11 @@ class File(QtCore.QObject, Item):
         else:
             self.clear_errors()
             self.state = self.NORMAL
-            self._copy_loaded_metadata(result)
+            postprocessors = []
+            if config.setting["guess_tracknumber_and_title"]:
+                postprocessors.append(self._guess_tracknumber_and_title)
+            self._copy_loaded_metadata(result, postprocessors)
         # use cached fingerprint from file metadata
-        config = get_config()
         if not config.setting["ignore_existing_acoustid_fingerprints"]:
             fingerprints = self.metadata.getall('acoustid_fingerprint')
             if fingerprints:
@@ -225,24 +278,25 @@ class File(QtCore.QObject, Item):
         self.update()
         callback(self)
 
-    def _copy_loaded_metadata(self, metadata):
-        filename, _ = os.path.splitext(self.base_filename)
+    def _copy_loaded_metadata(self, metadata, postprocessors=None):
         metadata['~length'] = format_time(metadata.length)
-        self.orig_metadata.copy(metadata)
-        if 'tracknumber' not in metadata:
-            tracknumber = tracknum_from_filename(self.base_filename)
-            if tracknumber is not None:
-                tracknumber = str(tracknumber)
-                metadata['tracknumber'] = tracknumber
-                if 'title' not in metadata:
-                    stripped_filename = filename.lstrip('0')
-                    tnlen = len(tracknumber)
-                    if stripped_filename[:tnlen] == tracknumber:
-                        metadata['title'] = stripped_filename[tnlen:].lstrip()
-        if 'title' not in metadata:
-            metadata['title'] = filename
+        if postprocessors:
+            for processor in postprocessors:
+                processor(metadata)
+        self.orig_metadata = metadata
         self.metadata.copy(metadata)
 
+    def _guess_tracknumber_and_title(self, metadata):
+        missing = {'tracknumber', 'title'} - set(metadata)
+        if missing:
+            guessed = tracknum_and_title_from_filename(self.base_filename)
+            for m in missing:
+                metadata[m] = getattr(guessed, m)
+
+    def _copy_file_info_tags(self, to_metadata, from_metadata):
+        for info in self.FILE_INFO_TAGS:
+            to_metadata[info] = from_metadata[info]
+
     def copy_metadata(self, metadata, preserve_deleted=True):
         acoustid = self.metadata["acoustid_id"]
         saved_metadata = {}
@@ -254,13 +308,11 @@ class File(QtCore.QObject, Item):
         deleted_tags = self.metadata.deleted_tags
         images_changed = self.metadata.images != metadata.images
         self.metadata.copy(metadata)
-        for info in FILE_INFO_TAGS:
-            metadata[info] = self.orig_metadata[info]
+        self._copy_file_info_tags(metadata, self.orig_metadata)
         if preserve_deleted:
             for tag in deleted_tags:
                 del self.metadata[tag]
-        for tag, values in saved_metadata.items():
-            self.metadata[tag] = values
+        self.metadata.update(saved_metadata)
 
         if acoustid and "acoustid_id" not in metadata.deleted_tags:
             self.metadata["acoustid_id"] = acoustid
@@ -283,7 +335,6 @@ class File(QtCore.QObject, Item):
         thread.run_task(
             partial(self._save_and_rename, self.filename, metadata),
             self._saving_finished,
-            priority=2,
             thread_pool=self.tagger.save_thread_pool)
 
     def _preserve_times(self, filename, func):
@@ -329,10 +380,9 @@ class File(QtCore.QObject, Item):
                 save()
         # Rename files
         if config.setting["rename_files"] or config.setting["move_files"]:
-            new_filename = self._rename(old_filename, metadata)
+            new_filename = self._rename(old_filename, metadata, config.setting)
         # Move extra files (images, playlists, etc.)
-        if config.setting["move_files"] and config.setting["move_additional_files"]:
-            self._move_additional_files(old_filename, new_filename)
+        self._move_additional_files(old_filename, new_filename, config)
         # Delete empty directories
         if config.setting["delete_empty_dirs"]:
             dirname = os.path.dirname(old_filename)
@@ -361,30 +411,28 @@ class File(QtCore.QObject, Item):
             return
         old_filename = new_filename = self.filename
         if error is not None:
-            self.state = File.ERROR
-            self.error_append(str(error))
+            self._set_error(error)
         else:
             self.filename = new_filename = result
             self.base_filename = os.path.basename(new_filename)
             length = self.orig_metadata.length
             temp_info = {}
-            for info in FILE_INFO_TAGS:
-                temp_info[info] = self.orig_metadata[info]
-            images_changed = self.orig_metadata.images != self.new_metadata.images
-            # Data is copied from New to Original because New may be
-            # a subclass to handle id3v23
+            self._copy_file_info_tags(temp_info, self.orig_metadata)
+            images_changed = self.orig_metadata.images != self.metadata.images
+            # Copy new metadata to original metadata, applying format specific
+            # conversions (e.g. for ID3v2.3)
             config = get_config()
+            new_metadata = self._format_specific_copy(self.metadata, config.setting)
             if config.setting["clear_existing_tags"]:
-                self.orig_metadata.copy(self.new_metadata)
+                self.orig_metadata = new_metadata
             else:
-                self.orig_metadata.update(self.new_metadata)
+                self.orig_metadata.update(new_metadata)
             # After saving deleted tags should no longer be marked deleted
             self.metadata.clear_deleted()
             self.orig_metadata.clear_deleted()
             self.orig_metadata.length = length
             self.orig_metadata['~length'] = format_time(length)
-            for k, v in temp_info.items():
-                self.orig_metadata[k] = v
+            self.orig_metadata.update(temp_info)
             self.clear_errors()
             self.clear_pending(signal=False)
             self._add_path_to_metadata(self.orig_metadata)
@@ -414,13 +462,16 @@ class File(QtCore.QObject, Item):
             settings = config.setting
         metadata = Metadata()
         if settings["clear_existing_tags"]:
-            metadata.copy(file_metadata)
+            # script_to_filename_with_metadata guarantees this is not modified
+            metadata = file_metadata
         else:
             metadata.copy(self.orig_metadata)
             metadata.update(file_metadata)
         (filename, new_metadata) = script_to_filename_with_metadata(
             naming_format, metadata, file=self, settings=settings)
-        # NOTE: the script_to_filename strips the extension away
+        if not filename:
+            return None
+        # NOTE: the filename generated by the naming script does not have a file extension
         ext = new_metadata.get('~extension', file_extension)
         return filename + '.' + ext.lstrip('.')
 
@@ -433,78 +484,70 @@ class File(QtCore.QObject, Item):
             new_filename = ''
         return new_filename, ext
 
-    def _format_filename(self, new_dirname, new_filename, metadata, settings):
+    def _clean_file_extension(self, filename):
+        """Takes a filename and converts the extension to lowercase.
+
+        If the file has no extension a default extension for the format is used.
+
+        Args:
+            filename: The filename
+
+        Returns: A tuple containing the filename with fixed extension and the extension itself.
+        """
+        filename, ext = self._fixed_splitext(filename)
+        if not ext and self.EXTENSIONS:
+            ext = self.EXTENSIONS[0]
+        ext = ext.lower()
+        return (filename + ext, ext)
+
+    def _format_filename(self, new_dirname, new_filename, metadata, settings, naming_format):
         old_filename = new_filename
-        new_filename, ext = self._fixed_splitext(new_filename)
-        ext = ext.lower()
-        new_filename = new_filename + ext
+        new_filename, ext = self._clean_file_extension(new_filename)
 
-        # expand the naming format
-        naming_format = settings['file_naming_format']
         if naming_format:
             new_filename = self._script_to_filename(naming_format, metadata, ext, settings)
+            if not new_filename:
+                new_filename = old_filename
             if not settings['rename_files']:
                 new_filename = os.path.join(os.path.dirname(new_filename), old_filename)
             if not settings['move_files']:
                 new_filename = os.path.basename(new_filename)
             win_compat = IS_WIN or settings['windows_compatibility']
-            new_filename = make_short_filename(new_dirname, new_filename,
-                                               win_compat)
-            # TODO: move following logic under util.filenaming
-            # (and reconsider its necessity)
-            # win32 compatibility fixes
-            if win_compat:
-                new_filename = new_filename.replace('./', '_/').replace('.\\', '_\\')
-            # replace . at the beginning of file and directory names
-            new_filename = new_filename.replace('/.', '/_').replace('\\.', '\\_')
-            if new_filename.startswith('.'):
-                new_filename = '_' + new_filename[1:]
-            # Fix for precomposed characters on OSX
-            if IS_MACOS:
-                new_filename = unicodedata.normalize("NFD", new_filename)
+            win_shorten_path = win_compat and not settings['windows_long_paths']
+            new_filename = make_short_filename(new_dirname, new_filename, win_shorten_path=win_shorten_path)
+            new_filename = make_save_path(new_filename, win_compat=win_compat, mac_compat=IS_MACOS)
         return new_filename
 
-    def make_filename(self, filename, metadata, settings=None):
+    def make_filename(self, filename, metadata, settings=None, naming_format=None):
         """Constructs file name based on metadata and file naming formats."""
         if settings is None:
             config = get_config()
             settings = config.setting
+        if naming_format is None:
+            naming_format = get_file_naming_script(settings)
         if settings["move_files"]:
             new_dirname = settings["move_files_to"]
             if not is_absolute_path(new_dirname):
-                new_dirname = os.path.normpath(os.path.join(os.path.dirname(filename), new_dirname))
+                new_dirname = os.path.join(os.path.dirname(filename), new_dirname)
         else:
             new_dirname = os.path.dirname(filename)
-        try:
-            new_dirname = os.path.realpath(new_dirname)
-        except FileNotFoundError:
-            # os.path.realpath can fail if cwd does not exist and path is relative
-            pass
         new_filename = os.path.basename(filename)
 
         if settings["rename_files"] or settings["move_files"]:
-            new_filename = self._format_filename(new_dirname, new_filename, metadata, settings)
+            new_filename = self._format_filename(new_dirname, new_filename, metadata, settings, naming_format)
 
         new_path = os.path.join(new_dirname, new_filename)
-        return new_path
+        return normpath(new_path)
 
-    def _rename(self, old_filename, metadata):
-        new_filename, ext = os.path.splitext(
-            self.make_filename(old_filename, metadata))
-
-        if old_filename == new_filename + ext:
+    def _rename(self, old_filename, metadata, settings=None):
+        new_filename = self.make_filename(old_filename, metadata, settings)
+        if old_filename == new_filename:
             return old_filename
 
         new_dirname = os.path.dirname(new_filename)
         if not os.path.isdir(new_dirname):
             os.makedirs(new_dirname)
-        tmp_filename = new_filename
-        i = 1
-        while (os.path.exists(new_filename + ext)
-               and not samefile(old_filename, new_filename + ext)):
-            new_filename = "%s (%d)" % (tmp_filename, i)
-            i += 1
-        new_filename = new_filename + ext
+        new_filename = get_available_filename(new_filename, old_filename)
         log.debug("Moving file %r => %r", old_filename, new_filename)
         move_ensure_casing(old_filename, new_filename)
         return new_filename
@@ -513,7 +556,7 @@ class File(QtCore.QObject, Item):
         """Save the cover images to disk."""
         if not metadata.images:
             return
-        counters = defaultdict(lambda: 0)
+        counters = Counter()
         images = []
         config = get_config()
         if config.setting["save_only_one_front_image"]:
@@ -525,41 +568,41 @@ class File(QtCore.QObject, Item):
         for image in images:
             image.save(dirname, metadata, counters)
 
-    def _move_additional_files(self, old_filename, new_filename):
+    def _move_additional_files(self, old_filename, new_filename, config):
         """Move extra files, like images, playlists..."""
-        new_path = os.path.dirname(new_filename)
-        old_path = os.path.dirname(old_filename)
-        if new_path == old_path:
-            # skip, same directory, nothing to move
-            return
-        config = get_config()
-        patterns = config.setting["move_additional_files_pattern"]
-        pattern_regexes = set()
-        for pattern in patterns.split():
-            pattern = pattern.strip()
-            if not pattern:
-                continue
-            pattern_regex = re.compile(fnmatch.translate(pattern), re.IGNORECASE)
-            match_hidden = pattern.startswith('.')
-            pattern_regexes.add((pattern_regex, match_hidden))
-        if not pattern_regexes:
-            return
-        moves = set()
-        try:
-            # TODO: use with statement with python 3.6+
-            for entry in os.scandir(old_path):
-                is_hidden = entry.name.startswith('.')
-                for pattern_regex, match_hidden in pattern_regexes:
-                    if is_hidden and not match_hidden:
-                        continue
-                    if pattern_regex.match(entry.name):
-                        new_file_path = os.path.join(new_path, entry.name)
-                        moves.add((entry.path, new_file_path))
-                        break  # we are done with this file
-        except OSError as why:
-            log.error("Failed to scan %r: %s", old_path, why)
-            return
+        if config.setting["move_files"] and config.setting["move_additional_files"]:
+            new_path = os.path.dirname(new_filename)
+            old_path = os.path.dirname(old_filename)
+            if new_path != old_path:
+                patterns_string = config.setting["move_additional_files_pattern"]
+                patterns = self._compile_move_additional_files_pattern(patterns_string)
+                try:
+                    moves = self._get_additional_files_moves(old_path, new_path, patterns)
+                    self._apply_additional_files_moves(moves)
+                except OSError as why:
+                    log.error("Failed to scan %r: %s", old_path, why)
 
+    @staticmethod
+    def _compile_move_additional_files_pattern(patterns_string):
+        return {
+            (re.compile(fnmatch.translate(pattern), re.IGNORECASE), pattern.startswith('.'))
+            for pattern in set(patterns_string.lower().split())
+        }
+
+    def _get_additional_files_moves(self, old_path, new_path, patterns):
+        if patterns:
+            with os.scandir(old_path) as scan:
+                for entry in scan:
+                    is_hidden = entry.name.startswith('.')
+                    for pattern_regex, match_hidden in patterns:
+                        if is_hidden and not match_hidden:
+                            continue
+                        if pattern_regex.match(entry.name):
+                            new_file_path = os.path.join(new_path, entry.name)
+                            yield (entry.path, new_file_path)
+                            break  # we are done with this file
+
+    def _apply_additional_files_moves(self, moves):
         for old_file_path, new_file_path in moves:
             # FIXME we shouldn't do this from a thread!
             if self.tagger.files.get(decode_filename(old_file_path)):
@@ -580,7 +623,8 @@ class File(QtCore.QObject, Item):
         self.state = File.REMOVED
 
     def move(self, parent):
-        if parent != self.parent:
+        # To be able to move a file the target must implement add_file(file)
+        if hasattr(parent, 'add_file') and parent != self.parent:
             log.debug("Moving %r from %r to %r", self, self.parent, parent)
             self.clear_lookup_task()
             self.tagger._acoustid.stop_analyze(self)
@@ -592,6 +636,9 @@ class File(QtCore.QObject, Item):
             self.parent = parent
             self.parent.add_file(self, new_album=new_album)
             self.acoustid_update()
+            return True
+        else:
+            return False
 
     def _move(self, parent):
         if parent != self.parent:
@@ -611,10 +658,13 @@ class File(QtCore.QObject, Item):
             self.acoustid_length = length or self.metadata.length // 1000
             self.tagger.acoustidmanager.add(self, None)
             self.acoustid_update()
+        config = get_config()
+        if config.setting['save_acoustid_fingerprints']:
+            self.metadata['acoustid_fingerprint'] = fingerprint
 
     def acoustid_update(self):
         recording_id = None
-        if self.parent and hasattr(self.parent, 'orig_metadata'):
+        if self.parent and self.parent.can_link_fingerprint:
             recording_id = self.parent.orig_metadata['musicbrainz_recordingid']
             if not recording_id:
                 recording_id = self.metadata['musicbrainz_recordingid']
@@ -629,34 +679,44 @@ class File(QtCore.QObject, Item):
     def is_saved(self):
         return self.similarity == 1.0 and self.state == File.NORMAL
 
+    def _tags_to_update(self, ignored_tags):
+        for name in set(self.metadata) | set(self.orig_metadata):
+            if name.startswith('~'):
+                continue
+            if not self.supports_tag(name):
+                continue
+            if name in ignored_tags:
+                continue
+            yield name
+
     def update(self, signal=True):
-        new_metadata = self.new_metadata
-        names = set(new_metadata.keys())
-        names.update(self.orig_metadata.keys())
-        config = get_config()
-        clear_existing_tags = config.setting["clear_existing_tags"]
-        ignored_tags = config.setting["compare_ignore_tags"]
-        for name in names:
-            if (not name.startswith('~') and self.supports_tag(name)
-                and name not in ignored_tags):
-                new_values = new_metadata.getall(name)
-                if not (new_values or clear_existing_tags
-                        or name in new_metadata.deleted_tags):
+        if not (self.state == File.ERROR and self.errors):
+            config = get_config()
+            clear_existing_tags = config.setting["clear_existing_tags"]
+            ignored_tags = set(config.setting["compare_ignore_tags"])
+
+            for name in self._tags_to_update(ignored_tags):
+                new_values = self.format_specific_metadata(self.metadata, name, config.setting)
+                if not (
+                    new_values
+                    or clear_existing_tags
+                    or name in self.metadata.deleted_tags
+                ):
                     continue
                 orig_values = self.orig_metadata.getall(name)
                 if orig_values != new_values:
-                    self.similarity = self.orig_metadata.compare(new_metadata, ignored_tags)
+                    self.similarity = self.orig_metadata.compare(self.metadata, ignored_tags)
                     if self.state == File.NORMAL:
                         self.state = File.CHANGED
                     break
-        else:
-            if (self.metadata.images
-                    and self.orig_metadata.images != self.metadata.images):
-                self.state = File.CHANGED
             else:
                 self.similarity = 1.0
-                if self.state == File.CHANGED:
-                    self.state = File.NORMAL
+                if self.state in (File.CHANGED, File.NORMAL):
+                    if (self.metadata.images
+                        and self.orig_metadata.images != self.metadata.images):
+                        self.state = File.CHANGED
+                    else:
+                        self.state = File.NORMAL
         if signal:
             log.debug("Updating file %r", self)
             self.update_item()
@@ -730,6 +790,8 @@ class File(QtCore.QObject, Item):
         m = self.metadata
         if column == "title" and not m["title"]:
             return self.base_filename
+        elif column == "covercount":
+            return self.cover_art_description()
         return m[column]
 
     def _lookup_finished(self, lookuptype, document, http, error):
@@ -781,10 +843,10 @@ class File(QtCore.QObject, Item):
 
     def _match_to_track(self, tracks, threshold=0):
         # multiple matches -- calculate similarities to each of them
-        def candidates():
-            for track in tracks:
-                yield self.metadata.compare_to_track(track, self.comparison_weights)
-
+        candidates = (
+            self.metadata.compare_to_track(track, self.comparison_weights)
+            for track in tracks
+        )
         no_match = SimMatchTrack(similarity=-1, releasegroup=None, release=None, track=None)
         best_match = find_best_match(candidates, no_match)
 
@@ -813,6 +875,7 @@ class File(QtCore.QObject, Item):
         self.clear_lookup_task()
         metadata = self.metadata
         self.set_pending()
+        config = get_config()
         self.lookup_task = self.tagger.mb_api.find_tracks(
             partial(self._lookup_finished, File.LOOKUP_METADATA),
             track=metadata['title'],
@@ -822,7 +885,7 @@ class File(QtCore.QObject, Item):
             tracks=metadata['totaltracks'],
             qdur=str(metadata.length // 2000),
             isrc=metadata['isrc'],
-            limit=QUERY_LIMIT)
+            limit=config.setting['query_limit'])
 
     def clear_lookup_task(self):
         if self.lookup_task:
