@@ -4,21 +4,22 @@
 #
 # Copyright (C) 2006-2009, 2012 Lukáš Lalinský
 # Copyright (C) 2007 Javier Kohen
-# Copyright (C) 2008-2011, 2014-2015, 2018-2020 Philipp Wolfer
+# Copyright (C) 2008-2011, 2014-2015, 2018-2022 Philipp Wolfer
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2009 Nikolai Prokoschenko
 # Copyright (C) 2011-2012 Michael Wiencek
 # Copyright (C) 2012 Chad Wilson
 # Copyright (C) 2012 stephen
-# Copyright (C) 2012, 2014, 2017 Wieland Hoffmann
-# Copyright (C) 2013-2014, 2017-2020 Laurent Monin
-# Copyright (C) 2014, 2017 Sophist-UK
+# Copyright (C) 2012, 2014, 2017, 2021 Wieland Hoffmann
+# Copyright (C) 2013-2014, 2017-2021 Laurent Monin
+# Copyright (C) 2014, 2017, 2021 Sophist-UK
 # Copyright (C) 2016-2017 Sambhav Kothari
 # Copyright (C) 2016-2017 Ville Skyttä
 # Copyright (C) 2017-2018 Antonio Larrosa
 # Copyright (C) 2018 Calvin Walton
 # Copyright (C) 2018 virusMac
-# Copyright (C) 2020 Bob Swift
+# Copyright (C) 2020-2022 Bob Swift
+# Copyright (C) 2021 Adam James
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -34,6 +35,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
 from collections import namedtuple
 import datetime
 from functools import reduce
@@ -42,6 +44,7 @@ import operator
 import re
 import unicodedata
 
+from picard.const.countries import RELEASE_COUNTRIES
 from picard.metadata import MULTI_VALUED_JOINER
 from picard.script.parser import (
     MultiValue,
@@ -49,7 +52,10 @@ from picard.script.parser import (
     ScriptRuntimeError,
     normalize_tagname,
 )
-from picard.util import uniqify
+from picard.util import (
+    pattern_as_regex,
+    uniqify,
+)
 
 
 try:
@@ -398,7 +404,7 @@ For example `$unset(performer:*)` would unset all performer tags."""
 def func_unset(parser, name):
     name = normalize_tagname(name)
     # Allow wild-card unset for certain keys
-    if name in ('performer:*', 'comment:*', 'lyrics:*'):
+    if name in {'performer:*', 'comment:*', 'lyrics:*'}:
         name = name[:-1]
         for key in list(parser.context.keys()):
             if key.startswith(name):
@@ -486,20 +492,22 @@ def func_copy(parser, new, old):
 
 
 @script_function(documentation=N_(
-    """`$copymerge(new,old)`
+    """`$copymerge(new,old[,keep_duplicates])`
 
 Merges metadata from variable `old` into `new`, removing duplicates and
     appending to the end, so retaining the original ordering. Like `$copy`, this
     will also copy multi-valued variables without flattening them.
 
+If `keep_duplicates` is set, then the duplicates will not be removed from the result.
+
 _Since Picard 1.0_"""
 ))
-def func_copymerge(parser, new, old):
+def func_copymerge(parser, new, old, keep_duplicates=False):
     new = normalize_tagname(new)
     old = normalize_tagname(old)
     newvals = parser.context.getall(new)
     oldvals = parser.context.getall(old)
-    parser.context[new] = uniqify(newvals + oldvals)
+    parser.context[new] = newvals + oldvals if keep_duplicates else uniqify(newvals + oldvals)
     return ""
 
 
@@ -758,14 +766,23 @@ def func_lenmulti(parser, multi, separator=MULTI_VALUED_JOINER):
     """`$performer(pattern="",join=", ")`
 
 Returns the performers where the performance type (e.g. "vocal") matches `pattern`, joined by `join`.
+You can specify a regular expression in the format `/pattern/flags`. `flags` are optional. Currently
+the only supported flag is "i" (ignore case). For example `$performer(/^guitars?$/i)` matches the
+performance type "guitar" or "Guitars", but not e.g. "bass guitar".
 
 _Since Picard 0.10_"""
 ))
 def func_performer(parser, pattern="", join=", "):
     values = []
+    try:
+        regex = pattern_as_regex(pattern, allow_wildcards=False)
+    except re.error:
+        return ''
     for name, value in parser.context.items():
-        if name.startswith("performer:") and pattern in name:
-            values.append(value)
+        if name.startswith("performer:"):
+            name, performance = name.split(':', 1)
+            if regex.search(performance):
+                values.append(value)
     return join.join(values)
 
 
@@ -1106,7 +1123,7 @@ def func_reverse(parser, text):
 
 
 @script_function(documentation=N_(
-    """`$substr(text,start,end)`
+    """`$substr(text,start[,end])`
 
 Returns the substring beginning with the character at the `start` index, up to
     (but not including) the character at the `end` index. Indexes are
@@ -1114,7 +1131,7 @@ Returns the substring beginning with the character at the `start` index, up to
     string. If the `start` or `end` indexes are left blank, they will default to
     the start and end of the string respectively."""
 ))
-def func_substr(parser, text, start_index, end_index):
+def func_substr(parser, text, start_index, end_index=None):
     try:
         start = int(start_index) if start_index else None
     except ValueError:
@@ -1173,7 +1190,7 @@ def func_foreach(parser, multi, loop_code, separator=MULTI_VALUED_JOINER):
 Standard 'while' loop. Executes `code` repeatedly until `condition` no longer
     evaluates to `True`. For each loop, the count is stored in the tag
     `_loop_count`. This allows the count value to be accessed within the `code`
-    script. The function limites the maximum number of iterations to 1000 as a
+    script. The function limits the maximum number of iterations to 1000 as a
     safeguard against accidentally creating an infinite loop."""
 ))
 def func_while(parser, condition, loop_code):
@@ -1243,12 +1260,13 @@ Returns a multi-value variable containing the elements between the `start` and
 The following example will create a multi-value variable with all artists
     in `%artists%` except the first, which can be used to create a "feat." list.
 
-Example:
+Examples:
 
-    $setmulti(supporting_artists,$slice(%artists%,1,))
+    $setmulti(supporting_artists,$slice(%artists%,1))
+    $setmulti(supporting_artists,$slice(%artists%,1,-1))
 """
 ))
-def func_slice(parser, multi, start_index, end_index, separator=MULTI_VALUED_JOINER):
+def func_slice(parser, multi, start_index, end_index=None, separator=MULTI_VALUED_JOINER):
     try:
         start = int(start_index.eval(parser)) if start_index else None
     except ValueError:
@@ -1352,3 +1370,150 @@ def func_unique(parser, multi, case_sensitive="", separator=MULTI_VALUED_JOINER)
     if not case_sensitive:
         multi_value._multi = list({v.lower(): v for v in multi_value}.values())
     return multi_value.separator.join(sorted(set(multi_value)))
+
+
+@script_function(documentation=N_(
+    """`$countryname(country_code,translate="")`
+
+Returns the name of the country for the specified country code.  If the country code is invalid an empty string will be returned.
+If translate is not blank, the output will be translated into the current locale language.
+"""
+))
+def func_countryname(parser, country_code, translate=""):
+    name = RELEASE_COUNTRIES.get(country_code.strip().upper(), "")
+    if translate:
+        return gettext_countries(name)
+    return name
+
+
+DateTuple = namedtuple('DateTuple', ('year', 'month', 'day'))
+
+
+def _split_date(date_to_parse, date_order="ymd"):
+    """Split the specified date into parts.
+
+    Args:
+        date_to_parse (str): Date string to parse
+        date_order (str, optional): Order of date elements. Can be "ymd", "mdy" or "dmy". Defaults to "ymd".
+
+    Returns:
+        tuple: Tuple of the date parts as (year, month, day)
+    """
+
+    parts = re.split(r'\D+', date_to_parse.strip())
+    parts.extend(['', '', ''])
+    date_order = date_order.lower()
+    if date_order == 'dmy':
+        return DateTuple(parts[2], parts[1], parts[0])
+    elif date_order == 'mdy':
+        return DateTuple(parts[2], parts[0], parts[1])
+    else:
+        return DateTuple(parts[0], parts[1], parts[2])
+
+
+@script_function(documentation=N_(
+    """`$year(date,date_order="ymd")`
+
+Returns the year portion of the specified date.  The default order is "ymd".  This can be changed by specifying
+either "dmy" or "mdy".  If the date is invalid an empty string will be returned.
+
+_Since Picard 2.7_"""
+))
+def func_year(parser, date_to_parse, date_order='ymd'):
+    return _split_date(date_to_parse, date_order).year
+
+
+@script_function(documentation=N_(
+    """`$month(date,date_order="ymd")`
+
+Returns the month portion of the specified date.  The default order is "ymd".  This can be changed by specifying
+either "dmy" or "mdy".  If the date is invalid an empty string will be returned.
+
+_Since Picard 2.7_"""
+))
+def func_month(parser, date_to_parse, date_order='ymd'):
+    return _split_date(date_to_parse, date_order).month
+
+
+@script_function(documentation=N_(
+    """`$day(date,date_order="ymd")`
+
+Returns the day portion of the specified date.  The default order is "ymd".  This can be changed by specifying
+either "dmy" or "mdy".  If the date is invalid an empty string will be returned.
+
+_Since Picard 2.7_"""
+))
+def func_day(parser, date_to_parse, date_order='ymd'):
+    return _split_date(date_to_parse, date_order).day
+
+
+@script_function(documentation=N_(
+    """`$dateformat(date,format="%Y-%m-%d",date_order="ymd")`
+
+Returns the input date in the specified `format`, which is based on the standard
+    Python `strftime` [format codes](https://strftime.org/). If no `format` is
+    specified the date will be returned in the form `2020-02-05`.  If the date or
+    format are invalid an empty string will be returned.
+
+    The default order for the input date is "ymd".  This can be changed by specifying
+    either "dmy" or "mdy".
+Note: Platform-specific formatting codes should be avoided to help ensure the
+    portability of scripts across the different platforms.  These codes include:
+    remove zero-padding (e.g. `%-d` and `%-m` on Linux or macOS, and their
+    equivalent `%#d` and `%#m` on Windows); element length specifiers (e.g.
+    `%3Y`); and hanging '%' at the end of the format string.
+
+_Since Picard 2.7_"""
+))
+def func_dateformat(parser, date_to_parse, date_format=None, date_order='ymd'):
+    # Handle case where format evaluates to ''
+    if not date_format:
+        date_format = '%Y-%m-%d'
+    yr, mo, da = _split_date(date_to_parse, date_order)
+    try:
+        date_object = datetime.date(int(yr), int(mo), int(da))
+    except ValueError:
+        return ''
+    try:
+        return date_object.strftime(date_format)
+    except ValueError:
+        return ''
+
+
+@script_function(eval_args=False, documentation=N_(
+    """`$is_multi(name)`
+
+Returns '1' if the argument is a multi-value tag and there are more than one elements, otherwise an empty string.
+
+Example:
+
+    $is_multi(%artists%)
+
+Result: 1 if there is more than one artist, otherwise "".
+
+_Since Picard 2.7_"""
+))
+def func_is_multi(parser, multi):
+    multi_value = MultiValue(parser, multi, MULTI_VALUED_JOINER)
+    return '' if len(multi_value) < 2 else '1'
+
+
+@script_function(eval_args=True, documentation=N_(
+    """`$cleanmulti(name)`
+
+Removes all empty string elements from the multi-value variable.
+
+Example:
+
+    $setmulti(test,one; ; two; three)
+    $cleanmulti(test)
+
+Result: Sets the value of 'test' to ["one", "two", "three"].
+
+_Since Picard 2.8_"""
+))
+def func_cleanmulti(parser, multi):
+    name = normalize_tagname(multi)
+    values = [str(value) for value in parser.context.getall(name) if value or value == 0]
+    parser.context[name] = values
+    return ""
